/*-
 * Copyright (C) @BABOLO  2008 Feb 10
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2008 Feb 10\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: db_pgsql.c,v 1.13 2008/04/19 00:09:38 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <errno.h>
#include <poll.h>
#include <mife.h>
#include <err.h>
#include <libpq-fe.h>
#include <babolo/BLINflag.h>
#include "pgoblin.h"
#include "pgob.h"

#define PRINTNOTNULL(A) (!(A) ? "" : (A))
#define PRINTANDNULL(A) PRINTNOTNULL(A), ((A) ? "~" : "")
#define PQSTATNDNULL(A) PRINTANDNULL(PQparameterStatus(db->conn, (A)))
#define ALIGN8(A)  popt[i].A, (strlen(popt[i].A) < 8 ? "		" : "	")
#define ALIGN16(A) popt[i].A                                                                          \
      , ((strlen(popt[i].A) < 8) ? "		" : (strlen(popt[i].envvar) < 16) ? "	" : " ")

#define PQEXEC(RESULT, QUERY, CONDITION, CCODE, RCONN) {                                              \
    ifBLIN_QV5(db->flags) fprintf(stderr, "PQexec: %s\n", (char*)(QUERY));                            \
    (RESULT) = PQexec(RCONN, (QUERY));                                                                \
    if  (!(RESULT)) {                                                                                 \
        ifBLIN_QV1(db->flags) warnx("result is absent - fatal error.\n");                             \
        return(CCODE);                                                                                \
    }                                                                                                 \
    ifBLIN_QV5(db->flags) fprintf(stderr, "PQresultStatus %d\n", PQresultStatus(RESULT));             \
    if  (PQresultStatus(RESULT) != (CONDITION)) {                                                     \
        if  (PQresultStatus(RESULT) == PGRES_FATAL_ERROR) {                                           \
            ifBLIN_QV1(db->flags)                                                                     \
                warnx("PQexec failed: %s", PQresultErrorMessage(RESULT));                             \
            return(CCODE);                                                                            \
        } else if ((CONDITION) == PGRES_COMMAND_OK && PQresultStatus(RESULT) == PGRES_TUPLES_OK) {    \
            ifBLIN_QV2(db->flags)                                                                     \
                warnx("PQexec wait for COMMAND_OK but received TUPLES_OK");                           \
        } else ifBLIN_QV1(db->flags) {                                                                \
            warnx("PQexec wait for %d but received %d", (CONDITION), PQresultStatus(RESULT));         \
            return(CCODE);                                                                            \
}   }   }

static void
NoticeProcessor(void *db, const char *message) {
    ifBLIN_QV1(((pgoblin_conn *)db)->flags) fprintf(stderr, "%s", message);
}

static int
pgoblin_shurecon(pgoblin_conn *db) {
    PQconninfoOption *popt;
    int i, ex = EX_OK;
    PGresult *res;
#   define DBSET (pgoblin.dbasetable[db->flags & PGOBLIN_DB_TYPE])

    if  (db->conn != NULL && PQstatus(db->conn) == CONNECTION_OK) return(0);
    ifBLIN_QV5(db->flags) {
        fprintf( stderr
               , "Connection defaults:\n"
                 "size=chr keyword	envvar		compiled val		label\n"
               );
        popt = PQconndefaults();
        for (i = 0; popt[i].keyword; i++) {
            fprintf( stderr
                   , "%4d=%s~	%s%s%s%s%s%s%s%s%s\n"
                   , popt[i].dispsize
                   , popt[i].dispchar
                   , ALIGN8(keyword)
                   , ALIGN16(envvar)
                   , PRINTNOTNULL(popt[i].compiled)
                   , !(popt[i].compiled) ? "	" : strlen(popt[i].compiled) < 7 ? "~	" : "~ "
                   , PRINTNOTNULL(popt[i].val)
                   , !(popt[i].val)
                   ? "		"
                   : strlen(popt[i].val) < 7 ? "~		" : "~	"
                   , popt[i].label
                   );
        }
        PQconninfoFree(popt);
        fprintf( stderr
               , "PGREALM=%s%s PGDATESTYLE=%s%s PGTZ=%s%s PGCLIENTENCODING=%s%s PGGEQO=%s%s\n"
               , PRINTANDNULL(getenv("PGREALM"))
               , PRINTANDNULL(getenv("PGDATESTYLE"))
               , PRINTANDNULL(getenv("PGTZ"))
               , PRINTANDNULL(getenv("PGCLIENTENCODING"))
               , PRINTANDNULL(getenv("PGGEQO"))
               );
    }
    /** R_CONN + **/
    ;;  if  (db->flags & PGOBLIN_PARMIN) {
    ;;      db->conn = PQconnectdb(db->dbname);
    ;;  } else {
    ;;      db->conn = PQsetdbLogin(db->host, db->port, NULL, NULL, db->dbname, db->username, NULL);
    ;;  }
    ;;  db->flags |= PGOBLIN_CLOSECONN;
    /** R_CONN - **/
    ifBLIN_QV5(db->flags) {
        fprintf(stderr, "PQoptions=%s~\n", PQoptions(db->conn));
        fprintf( stderr
               , "PQdb=%s~ PQuser=%s~ PQpass=%s%s "
                 "PQhost=%s%s PQport=%s%s "
                 "PQtty=%s%s\n"
               , PQdb(db->conn), PQuser(db->conn)
               , PRINTANDNULL(PQpass(db->conn))
               , PRINTANDNULL(PQhost(db->conn))
               , PRINTANDNULL(PQport(db->conn))
               , PRINTANDNULL(PQtty(db->conn))
               );
        fprintf( stderr
               , "server_version=%s%s "
                 "client_encoding=%s%s "
                 "is_superuser=%s%s "
                 "session_authorization=%s%s "
                 "DateStyle=%s%s\n"
               , PQSTATNDNULL("server_version")
               , PQSTATNDNULL("client_encoding")
               , PQSTATNDNULL("is_superuser")
               , PQSTATNDNULL("session_authorization")
               , PQSTATNDNULL("DateStyle")
               );
        fprintf( stderr
               , "PQprotocolVersion=%d PQsocket=%d PQbackendPID=%d\n"
               , PQprotocolVersion(db->conn)
               , PQsocket(db->conn)
               , PQbackendPID(db->conn)
               );
        fprintf(stderr, "PQstatus %d\n", PQstatus(db->conn));
    }
    PQsetNoticeProcessor(db->conn, NoticeProcessor, db);
    if  (PQstatus(db->conn) != CONNECTION_BAD) {
        if  (  db->flags & PGOBLIN_NEEDINTIME
            && !(db->flags & (PGOBLIN_INT64TIME | PGOBLIN_FLOATTIME))
            ) {
            if  (!db->intran) {
/*          if  (PQtransactionStatus(db->conn) != PQTRANS_INTRANS) { */
                PQEXEC(res, "BEGIN", PGRES_COMMAND_OK, -1, db->conn);
                PQclear(res);
            }
            PQEXEC( res
                  , "DECLARE pgoblin BINARY NO SCROLL CURSOR FOR "
                    "SELECT '2000-01-01 00:00:01'::timestamp;"
                    "FETCH 1 FROM pgoblin"
                  , PGRES_TUPLES_OK, -1
                  , db->conn
                  );
            if  (PQgetisnull(res, 0, 0) || (PQgetlength(res, 0, 0) < 8)) {
                ex = EX_DATAERR;
                errno = EINVAL;
            } else {
                if  (*((u_char*)(PQgetvalue(res, 0, 0))) == 077) {
                    db->flags |= PGOBLIN_FLOATTIME;
                    ifBLIN_QV3(db->flags) fprintf(stderr, "shurecon float timestamp format\n");
                } else if (*((u_char*)(PQgetvalue(res, 0, 0))) == 0) {
                    db->flags |= PGOBLIN_INT64TIME;
                    ifBLIN_QV3(db->flags) fprintf(stderr, "shurecon int64 timestamp format\n");
                } else {
                    ifBLIN_QV1(db->flags) fprintf(stderr, "shurecon unknown timestamp format\n");
                    ex = EX_DATAERR;
                    errno = EFTYPE;
            }   }
            PQclear(res);
            PQEXEC(res, "CLOSE pgoblin", PGRES_COMMAND_OK, -1, db->conn);
            PQclear(res);
            if  (!db->intran) {
                PQEXEC(res, "END", PGRES_COMMAND_OK, -1, db->conn);
                PQclear(res);
        }   }
        return(0);
    }
    ifBLIN_QV1(db->flags) warnx("Connection to PostgreSQL failed: %s\n", PQerrorMessage(db->conn));
    return(1);
#   undef DBSET
}

static int
pgoblin_tarry(PGconn *conn, int timeout) {
    int           sock, i = 0;
    struct pollfd pfd;

    if  ((sock = PQsocket(conn)) < 0) {
        i = -2;
        errno = EINVAL;
    } else if (timeout != 0) {
        bzero(&pfd, sizeof(pfd));
        pfd.fd = sock;
        pfd.events = POLLIN | POLLRDNORM | POLLRDBAND | POLLPRI;
        i = poll(&pfd, 1, timeout);
    }
    if  (i >= 0) {
        if  (!PQconsumeInput(conn)) i = -1;
    }
    return(i);
}

#define U0 0
#define U1 1
#define U2 2
#define U3 3
#define U4 4

static u_char miss[][PGOBLIN_CR_MAX] =
/*  OK   TUPL OUT  IN   STR  BIN   */
{{  U2,  U2,  U2,  U2,  U2,  U2 } /* PGRES_EMPTY_QUERY    */
,{  U0,  U2,  U2,  U2,  U2,  U2 } /* PGRES_COMMAND_OK     */
,{  U1,  U0,  U2,  U2,  U2,  U2 } /* PGRES_TUPLES_OK      */
,{  U2,  U2,  U0,  U2,  U2,  U2 } /* PGRES_COPY_OUT       */
,{  U2,  U2,  U2,  U0,  U2,  U2 } /* PGRES_COPY_IN        */
,{  U3,  U3,  U3,  U3,  U3,  U3 } /* PGRES_BAD_RESPONSE   */
,{  U3,  U3,  U3,  U3,  U3,  U3 } /* PGRES_NONFATAL_ERROR */
,{  U4,  U4,  U4,  U4,  U4,  U4 } /* PGRES_FATAL_ERROR    */
};

static ExecStatusType es[] = {PGRES_COMMAND_OK, PGRES_TUPLES_OK, PGRES_COPY_OUT, PGRES_COPY_IN};

static int
pgoblin_query(pgoblin_conn *db, PGresult **rs, const char *query, u_char cond) {
    ExecStatusType st;
    PGresult *res;
    char *q, *p;
    int ex = 0;

    res = NULL;
    switch (cond) {
    case PGOBLIN_CR_OK:
    case PGOBLIN_CR_TUPL:
    case PGOBLIN_CR_OUT:
    case PGOBLIN_CR_IN:
        ifBLIN_QV5(db->flags) fprintf(stderr, "PQexec %d=%s~\n", cond, query);
        if  (!(res = PQexec(db->conn, query))) {
            ifBLIN_QV1(db->flags) warnx("PQ result is absent - fatal error.\n");
            ERROUT(EX_UNAVAILABLE, ENODEV);
        }
        st = PQresultStatus(res);
        ifBLIN_QV5(db->flags)
            fprintf(stderr, "PQresultStatus %d %s, diff %d\n", st, PQresStatus(st), miss[st][cond]);
        switch (miss[st][cond]) {
        case U1:
            ifBLIN_QV2(db->flags)
                warnx("PQexec waits for PGRES_COMMAND_OK but received PGRES_TUPLES_OK");
            break;
        case U2:
            ifBLIN_QV1(db->flags) warnx( "PQexec wait for %d %s but received %d %s"
                                       , es[cond], PQresStatus(es[cond])
                                       , st, PQresStatus(st)
                                       );
            ex = miss[st][cond];
            break;
        case U4:
        case U3:
            ifBLIN_QV1(db->flags) warnx("PQexec failed: %s", PQresultErrorMessage(res));
            ex = miss[st][cond];
            break;
        }
        break;
    case PGOBLIN_CR_SSEL:
        p = "pgoblin";
        goto rest;
    case PGOBLIN_CR_BSEL:
        p = "pgoblin BINARY";
    rest:
        if  (!db->intran && (ex = pgoblin_query(db, NULL, "BEGIN", PGOBLIN_CR_OK))) goto out;
        asprintf(&q, "DECLARE %s NO SCROLL CURSOR FOR %s", p, query);
        if  (q == NULL) {
            ifBLIN_QV1(db->flags) warnx("pgoblin_query: No memory");
            ERROUT(EX_UNAVAILABLE, ENOMEM);
        }
        ex = pgoblin_query(db, &res, q, PGOBLIN_CR_OK);
        free(q);
        break;
    default:
        ifBLIN_QV1(db->flags) warnx("pgoblin_query pgsql illegal cond=%d", cond);
        ex = -1;
    }
out:
    if  (rs) *rs = res; else if (res) PQclear(res);
    return(ex);
}

static char *
pgoblin_notifies(PGconn *conn) {
    PGnotify *n;
    char *c = NULL;

    n = PQnotifies(conn);
    if  (n && n->relname) c = strdup(n->relname);
    PQfreemem(n);
    return(c);
}

static void
pgoblin_free(void **c) {
    if  (c) {
        if  (*c) free(*c);
        *c = NULL;
}   }

static void
pgoblin_finish(PGconn **conn) {
    if  (conn) {
        if  (*conn) PQfinish(*conn);
        *conn = NULL;
}   }

static void
pgoblin_frees(PGresult **res) {
    if  (res) {
        if  (*res) PQclear(*res);
        *res = NULL;
}   }

static void
pgoblin_freemem(void **m) {
    if  (m) {
        if  (*m) PQfreemem(*m);
        *m = NULL;
}   }

static int
pgoblin_getstream(pgoblin_conn *db, PGresult **res, int quant, u_char cond) {
    int ex = EX_OK;
    char *query;

    *res = NULL;
    if  ((cond != PGOBLIN_CR_SSEL) && (cond != PGOBLIN_CR_BSEL)) {
        ifBLIN_QV1(db->flags) warnx("pgoblin_getstream: Illegal cond %d", cond);
        ERROUT(EX_USAGE, EINVAL);
    }
    if  (asprintf(&query, "FETCH %d FROM pgoblin", quant) < 0) {
        ifBLIN_QV1(db->flags) warnx("pgoblin_getstream: No memory");
        ERROUT(EX_OSERR, ENOMEM);
    }
    ex = pgoblin_query(db, res, query, PGOBLIN_CR_TUPL);
    free(query);
out:
    return(ex);
}

static int
pgoblin_getcopy(pgoblin_conn *db, void **buf) {
    return(PQgetCopyData(db->conn, (char**)buf, 0));
}

static int
pgoblin_putcopy(pgoblin_conn *db, const void *buf, int nby) {
    return(PQputCopyData(db->conn, buf, nby));
}

static int
pgoblin_endstream(pgoblin_conn *db, u_char cond) {
    ExecStatusType st;
    PGresult *res;
    int ex = EX_OK;

    switch (cond) {
    case PGOBLIN_CR_IN:
        if  (PQputCopyEnd(db->conn, NULL) < 1) {
            ifBLIN_QV1(db->flags)
                warnx("pgoblin_copyin: endcopy failed: %s", PQerrorMessage(db->conn));
            ERROUT(-EX_DATAERR, EINVAL);
        }
        /* fallthrou */
    case PGOBLIN_CR_OUT:
        while ((res = PQgetResult(db->conn))) {
            st = PQresultStatus(res);
            ifBLIN_QV5(db->flags) fprintf(stderr, "PQresultStatus %d %s\n", st, PQresStatus(st));
            if  (st != PGRES_COMMAND_OK) {
                ifBLIN_QV1(db->flags) warnx("pgoblin_endstream pgsql: %s", PQresultErrorMessage(res));
                ex = EX_SOFTWARE;
                break;
            }
            PQclear(res);
            res = NULL;
        }
        break;
    case PGOBLIN_CR_SSEL:
    case PGOBLIN_CR_BSEL:
        ex = pgoblin_query(db, NULL, "CLOSE pgoblin", PGOBLIN_CR_OK);
        if  (!db->intran) {
            ex |= pgoblin_query(db, &res, "END", PGOBLIN_CR_OK);
            PQclear(res);
        }
        break;
    default:
        ifBLIN_QV1(db->flags) warnx("pgoblin_endstream pgsql %d", cond);
        ex = EX_USAGE;
    }
out:
    if  (ex) errno = EINVAL;
    return(ex);
}

pgoblin_dbases pgoblin_db_pgsql =
{ "pgsql"
,                                      pgoblin_shurecon                            /* shurecon  */
, (int  (*)(pgoblin_conn*, void**, const char*, u_char))pgoblin_query              /* exec      */
, (int  (*)(void*))                    PQntuples                                   /* ntuples   */
, (int  (*)(void*))                    PQnfields                                   /* nfields   */
, (int  (*)(void*, int, int))          PQgetisnull                                 /* getisnull */
, (int  (*)(void*, int, int))          PQgetlength                                 /* getlength */
, (char*(*)(void*, int, int))          PQgetvalue                                  /* getvalue  */
, (int  (*)(pgoblin_conn*, void**, int, u_char))pgoblin_getstream                  /* getstream */
,                                      pgoblin_getcopy                             /* getcopy   */
,                                      pgoblin_putcopy                             /* putcopy   */
,                                      pgoblin_endstream                           /* endstream */
, (int  (*)(void*, int))               pgoblin_tarry                               /* tarry     */
, (char*(*)(void*))                    pgoblin_notifies                            /* notifies  */
, (void (*)(void**))                   pgoblin_freemem                             /* freecopy  */
, (void (*)(char**))                   pgoblin_free                                /* freenote  */
, (void (*)(void**))                   pgoblin_frees                               /* clear     */
, (void (*)(void**))                   pgoblin_finish                              /* finish    */
};
